@@ -8,13 +8,18 @@
 import asyncio
 import functools
 import inspect
+import logging
 import time
+import warnings
 
 from . import connection
 from . import connect_utils
 from . import exceptions
 
 
+logger = logging.getLogger(__name__)
+
+
 class PoolConnectionProxyMeta(type):
 
     def __new__(mcls, name, bases, dct, *, wrap=False):
@@ -74,8 +79,7 @@ class PoolConnectionProxy(connection._ConnectionProxy,
 
     def _detach(self) -> connection.Connection:
         if self._con is None:
-            raise exceptions.InterfaceError(
-                'cannot detach PoolConnectionProxy: already detached')
+            return
 
         con, self._con = self._con, None
         con._set_proxy(None)
@@ -92,67 +96,50 @@ class PoolConnectionProxy(connection._ConnectionProxy,
 
 class PoolConnectionHolder:
 
-    __slots__ = ('_con', '_pool', '_loop',
-                 '_connect_args', '_connect_kwargs',
-                 '_max_queries', '_setup', '_init',
+    __slots__ = ('_con', '_pool', '_loop', '_proxy',
+                 '_max_queries', '_setup',
                  '_max_inactive_time', '_in_use',
-                 '_inactive_callback', '_timeout')
+                 '_inactive_callback', '_timeout',
+                 '_generation')
 
-    def __init__(self, pool, *, connect_args, connect_kwargs,
-                 max_queries, setup, init, max_inactive_time):
+    def __init__(self, pool, *, max_queries, setup, max_inactive_time):
 
         self._pool = pool
         self._con = None
+        self._proxy = None
 
-        self._connect_args = connect_args
-        self._connect_kwargs = connect_kwargs
         self._max_queries = max_queries
         self._max_inactive_time = max_inactive_time
         self._setup = setup
-        self._init = init
         self._inactive_callback = None
-        self._in_use = False
+        self._in_use = None  # type: asyncio.Future
         self._timeout = None
+        self._generation = None
 
     async def connect(self):
-        assert self._con is None
+        if self._con is not None:
+            raise exceptions.InternalClientError(
+                'PoolConnectionHolder.connect() called while another '
+                'connection already exists')
 
-        if self._pool._working_addr is None:
-            # First connection attempt on this pool.
-            con = await connection.connect(
-                *self._connect_args,
-                loop=self._pool._loop,
-                connection_class=self._pool._connection_class,
-                **self._connect_kwargs)
-
-            self._pool._working_addr = con._addr
-            self._pool._working_config = con._config
-            self._pool._working_params = con._params
-
-        else:
-            # We've connected before and have a resolved address,
-            # and parsed options and config.
-            con = await connect_utils._connect_addr(
-                loop=self._pool._loop,
-                addr=self._pool._working_addr,
-                timeout=self._pool._working_params.connect_timeout,
-                config=self._pool._working_config,
-                params=self._pool._working_params,
-                connection_class=self._pool._connection_class)
-
-        if self._init is not None:
-            await self._init(con)
-
-        self._con = con
+        self._con = await self._pool._get_new_connection()
+        self._generation = self._pool._generation
 
     async def acquire(self) -> PoolConnectionProxy:
         if self._con is None or self._con.is_closed():
             self._con = None
             await self.connect()
 
+        elif self._generation != self._pool._generation:
+            # Connections have been expired, re-connect the holder.
+            self._pool._loop.create_task(
+                self._con.close(timeout=self._timeout))
+            self._con = None
+            await self.connect()
+
         self._maybe_cancel_inactive_callback()
 
-        proxy = PoolConnectionProxy(self, self._con)
+        self._proxy = proxy = PoolConnectionProxy(self, self._con)
 
         if self._setup is not None:
             try:
@@ -163,102 +150,146 @@ class PoolConnectionHolder:
                 # we close it.  A new connection will be created
                 # when `acquire` is called again.
                 try:
-                    proxy._detach()
-                    # Use `close` to close the connection gracefully.
+                    # Use `close()` to close the connection gracefully.
                     # An exception in `setup` isn't necessarily caused
-                    # by an IO or a protocol error.
+                    # by an IO or a protocol error.  close() will
+                    # do the necessary cleanup via _release_on_close().
                     await self._con.close()
                 finally:
-                    self._con = None
                     raise ex
 
-        self._in_use = True
+        self._in_use = self._pool._loop.create_future()
+
         return proxy
 
     async def release(self, timeout):
-        assert self._in_use
-        self._in_use = False
+        if self._in_use is None:
+            raise exceptions.InternalClientError(
+                'PoolConnectionHolder.release() called on '
+                'a free connection holder')
+
+        if self._con.is_closed():
+            # When closing, pool connections perform the necessary
+            # cleanup, so we don't have to do anything else here.
+            return
+
         self._timeout = None
 
-        if self._con.is_closed():
-            self._con = None
+        if self._con._protocol.queries_count >= self._max_queries:
+            # The connection has reached its maximum utilization limit,
+            # so close it.  Connection.close() will call _release().
+            await self._con.close(timeout=timeout)
+            return
 
-        elif self._con._protocol.queries_count >= self._max_queries:
+        if self._generation != self._pool._generation:
+            # The connection has expired because it belongs to
+            # an older generation (Pool.expire_connections() has
+            # been called.)
+            await self._con.close(timeout=timeout)
+            return
+
+        try:
+            budget = timeout
+
+            if self._con._protocol._is_cancelling():
+                # If the connection is in cancellation state,
+                # wait for the cancellation
+                started = time.monotonic()
+                await asyncio.wait_for(
+                    self._con._protocol._wait_for_cancellation(),
+                    budget, loop=self._pool._loop)
+                if budget is not None:
+                    budget -= time.monotonic() - started
+
+            await self._con.reset(timeout=budget)
+        except Exception as ex:
+            # If the `reset` call failed, terminate the connection.
+            # A new one will be created when `acquire` is called
+            # again.
             try:
-                await self._con.close(timeout=timeout)
+                # An exception in `reset` is most likely caused by
+                # an IO error, so terminate the connection.
+                self._con.terminate()
             finally:
-                self._con = None
+                raise ex
 
+        # Free this connection holder and invalidate the
+        # connection proxy.
+        self._release()
+
+        # Rearm the connection inactivity timer.
+        self._setup_inactive_callback()
+
+    async def wait_until_released(self):
+        if self._in_use is None:
+            return
         else:
-            try:
-                budget = timeout
-
-                if self._con._protocol._is_cancelling():
-                    # If the connection is in cancellation state,
-                    # wait for the cancellation
-                    started = time.monotonic()
-                    await asyncio.wait_for(
-                        self._con._protocol._wait_for_cancellation(),
-                        budget, loop=self._pool._loop)
-                    if budget is not None:
-                        budget -= time.monotonic() - started
-
-                await self._con.reset(timeout=budget)
-            except Exception as ex:
-                # If the `reset` call failed, terminate the connection.
-                # A new one will be created when `acquire` is called
-                # again.
-                try:
-                    # An exception in `reset` is most likely caused by
-                    # an IO error, so terminate the connection.
-                    self._con.terminate()
-                finally:
-                    self._con = None
-                    raise ex
-
-        assert self._inactive_callback is None
-        if self._max_inactive_time and self._con is not None:
-            self._inactive_callback = self._pool._loop.call_later(
-                self._max_inactive_time, self._deactivate_connection)
+            await self._in_use
 
     async def close(self):
-        self._maybe_cancel_inactive_callback()
-        if self._con is None:
-            return
-        if self._con.is_closed():
-            self._con = None
-            return
-
-        try:
+        if self._con is not None:
+            # Connection.close() will call _release_on_close() to
+            # finish holder cleanup.
             await self._con.close()
-        finally:
-            self._con = None
 
     def terminate(self):
-        self._maybe_cancel_inactive_callback()
-        if self._con is None:
-            return
-        if self._con.is_closed():
-            self._con = None
-            return
-
-        try:
+        if self._con is not None:
+            # Connection.terminate() will call _release_on_close() to
+            # finish holder cleanup.
             self._con.terminate()
-        finally:
-            self._con = None
+
+    def _setup_inactive_callback(self):
+        if self._inactive_callback is not None:
+            raise exceptions.InternalClientError(
+                'pool connection inactivity timer already exists')
+
+        if self._max_inactive_time:
+            self._inactive_callback = self._pool._loop.call_later(
+                self._max_inactive_time, self._deactivate_inactive_connection)
 
     def _maybe_cancel_inactive_callback(self):
         if self._inactive_callback is not None:
             self._inactive_callback.cancel()
             self._inactive_callback = None
 
-    def _deactivate_connection(self):
-        assert not self._in_use
-        if self._con is None or self._con.is_closed():
-            return
-        self._con.terminate()
+    def _deactivate_inactive_connection(self):
+        if self._in_use is not None:
+            raise exceptions.InternalClientError(
+                'attempting to deactivate an acquired connection')
+
+        if self._con is not None:
+            # The connection is idle and not in use, so it's fine to
+            # use terminate() instead of close().
+            self._con.terminate()
+            # Must call clear_connection, because _deactivate_connection
+            # is called when the connection is *not* checked out, and
+            # so terminate() above will not call the below.
+            self._release_on_close()
+
+    def _release_on_close(self):
+        self._maybe_cancel_inactive_callback()
+        self._release()
         self._con = None
 
+    def _release(self):
+        """Release this connection holder."""
+        if self._in_use is None:
+            # The holder is not checked out.
+            return
+
+        if not self._in_use.done():
+            self._in_use.set_result(None)
+        self._in_use = None
+
+        # Deinitialize the connection proxy.  All subsequent
+        # operations on it will fail.
+        if self._proxy is not None:
+            self._proxy._detach()
+            self._proxy = None
+
+        # Put ourselves back to the pool queue.
+        self._pool._queue.put_nowait(self)
+
 
 class Pool:
     """A connection pool.
@@ -272,9 +303,10 @@ class Pool:
     """
 
     __slots__ = ('_queue', '_loop', '_minsize', '_maxsize',
+                 '_init', '_connect_args', '_connect_kwargs',
                  '_working_addr', '_working_config', '_working_params',
-                 '_holders', '_initialized', '_closed',
-                 '_connection_class')
+                 '_holders', '_initialized', '_initializing', '_closing',
+                 '_closed', '_connection_class', '_generation')
 
     def __init__(self, *connect_args,
                  min_size,
@@ -287,6 +319,14 @@ class Pool:
                  connection_class,
                  **connect_kwargs):
 
+        if len(connect_args) > 1:
+            warnings.warn(
+                "Passing multiple positional arguments to asyncpg.Pool "
+                "constructor is deprecated and will be removed in "
+                "asyncpg 0.17.0.  The non-deprecated form is "
+                "asyncpg.Pool(<dsn>, **kwargs)",
+                DeprecationWarning, stacklevel=2)
+
         if loop is None:
             loop = asyncio.get_event_loop()
         self._loop = loop
@@ -309,11 +349,17 @@ class Pool:
                 'max_inactive_connection_lifetime is expected to be greater '
                 'or equal to zero')
 
+        if not issubclass(connection_class, connection.Connection):
+            raise TypeError(
+                'connection_class is expected to be a subclass of '
+                'asyncpg.Connection, got {!r}'.format(connection_class))
+
         self._minsize = min_size
         self._maxsize = max_size
 
         self._holders = []
         self._initialized = False
+        self._initializing = False
         self._queue = asyncio.LifoQueue(maxsize=self._maxsize, loop=self._loop)
 
         self._working_addr = None
@@ -322,17 +368,19 @@ class Pool:
 
         self._connection_class = connection_class
 
+        self._closing = False
         self._closed = False
+        self._generation = 0
+        self._init = init
+        self._connect_args = connect_args
+        self._connect_kwargs = connect_kwargs
 
         for _ in range(max_size):
             ch = PoolConnectionHolder(
                 self,
-                connect_args=connect_args,
-                connect_kwargs=connect_kwargs,
                 max_queries=max_queries,
                 max_inactive_time=max_inactive_connection_lifetime,
-                setup=setup,
-                init=init)
+                setup=setup)
 
             self._holders.append(ch)
             self._queue.put_nowait(ch)
@@ -340,9 +388,20 @@ class Pool:
     async def _async__init__(self):
         if self._initialized:
             return
+        if self._initializing:
+            raise exceptions.InterfaceError(
+                'pool is being initialized in another task')
         if self._closed:
             raise exceptions.InterfaceError('pool is closed')
+        self._initializing = True
+        try:
+            await self._initialize()
+            return self
+        finally:
+            self._initializing = False
+            self._initialized = True
 
+    async def _initialize(self):
         if self._minsize:
             # Since we use a LIFO queue, the first items in the queue will be
             # the last ones in `self._holders`.  We want to pre-connect the
@@ -365,8 +424,75 @@ class Pool:
 
                 await asyncio.gather(*connect_tasks, loop=self._loop)
 
-        self._initialized = True
-        return self
+    def set_connect_args(self, dsn=None, **connect_kwargs):
+        r"""Set the new connection arguments for this pool.
+
+        The new connection arguments will be used for all subsequent
+        new connection attempts.  Existing connections will remain until
+        they expire. Use :meth:`Pool.expire_connections()
+        <asyncpg.pool.Pool.expire_connections>` to expedite the connection
+        expiry.
+
+        :param str dsn:
+            Connection arguments specified using as a single string in
+            the following format:
+            ``postgres://user:pass@host:port/database?option=value``.
+
+        :param \*\*connect_kwargs:
+            Keyword arguments for the :func:`~asyncpg.connection.connect`
+            function.
+
+        .. versionadded:: 0.16.0
+        """
+
+        self._connect_args = [dsn]
+        self._connect_kwargs = connect_kwargs
+        self._working_addr = None
+        self._working_config = None
+        self._working_params = None
+
+    async def _get_new_connection(self):
+        if self._working_addr is None:
+            # First connection attempt on this pool.
+            con = await connection.connect(
+                *self._connect_args,
+                loop=self._loop,
+                connection_class=self._connection_class,
+                **self._connect_kwargs)
+
+            self._working_addr = con._addr
+            self._working_config = con._config
+            self._working_params = con._params
+
+        else:
+            # We've connected before and have a resolved address,
+            # and parsed options and config.
+            con = await connect_utils._connect_addr(
+                loop=self._loop,
+                addr=self._working_addr,
+                timeout=self._working_params.connect_timeout,
+                config=self._working_config,
+                params=self._working_params,
+                connection_class=self._connection_class)
+
+        if self._init is not None:
+            try:
+                await self._init(con)
+            except Exception as ex:
+                # If a user-defined `init` function fails, we don't
+                # know if the connection is safe for re-use, hence
+                # we close it.  A new connection will be created
+                # when `acquire` is called again.
+                try:
+                    # Use `close()` to close the connection gracefully.
+                    # An exception in `init` isn't necessarily caused
+                    # by an IO or a protocol error.  close() will
+                    # do the necessary cleanup via _release_on_close().
+                    await con.close()
+                finally:
+                    raise ex
+
+        return con
 
     async def execute(self, query: str, *args, timeout: float=None) -> str:
         """Execute an SQL command (or commands).
@@ -468,7 +594,10 @@ class Pool:
                 ch._timeout = timeout
                 return proxy
 
+        if self._closing:
+            raise exceptions.InterfaceError('pool is closing')
         self._check_init()
+
         if timeout is None:
             return await _acquire_impl()
         else:
@@ -488,14 +617,6 @@ class Pool:
         .. versionchanged:: 0.14.0
             Added the *timeout* parameter.
         """
-        async def _release_impl(ch: PoolConnectionHolder, timeout: float):
-            try:
-                await ch.release(timeout)
-            finally:
-                self._queue.put_nowait(ch)
-
-        self._check_init()
-
         if (type(connection) is not PoolConnectionProxy or
                 connection._holder._pool is not self):
             raise exceptions.InterfaceError(
@@ -507,38 +628,95 @@ class Pool:
             # Already released, do nothing.
             return
 
-        con = connection._detach()
-        con._on_release()
+        self._check_init()
 
+        # Let the connection do its internal housekeeping when its released.
+        connection._con._on_release()
+
+        ch = connection._holder
         if timeout is None:
-            timeout = connection._holder._timeout
+            timeout = ch._timeout
 
         # Use asyncio.shield() to guarantee that task cancellation
         # does not prevent the connection from being returned to the
         # pool properly.
-        return await asyncio.shield(
-            _release_impl(connection._holder, timeout), loop=self._loop)
+        return await asyncio.shield(ch.release(timeout), loop=self._loop)
 
     async def close(self):
-        """Gracefully close all connections in the pool."""
+        """Attempt to gracefully close all connections in the pool.
+
+        Wait until all pool connections are released, close them and
+        shut down the pool.  If any error (including cancellation) occurs
+        in ``close()`` the pool will terminate by calling
+        :meth:`Pool.terminate() <pool.Pool.terminate>`.
+
+        It is advisable to use :func:`python:asyncio.wait_for` to set
+        a timeout.
+
+        .. versionchanged:: 0.16.0
+            ``close()`` now waits until all pool connections are released
+            before closing them and the pool.  Errors raised in ``close()``
+            will cause immediate pool termination.
+        """
         if self._closed:
             return
         self._check_init()
-        self._closed = True
-        coros = [ch.close() for ch in self._holders]
-        await asyncio.gather(*coros, loop=self._loop)
+
+        self._closing = True
+
+        try:
+            warning_callback = self._loop.call_later(
+                60, self._warn_on_long_close)
+
+            release_coros = [
+                ch.wait_until_released() for ch in self._holders]
+            await asyncio.gather(*release_coros, loop=self._loop)
+
+            close_coros = [
+                ch.close() for ch in self._holders]
+            await asyncio.gather(*close_coros, loop=self._loop)
+
+        except Exception:
+            self.terminate()
+            raise
+
+        finally:
+            warning_callback.cancel()
+            self._closed = True
+            self._closing = False
+
+    def _warn_on_long_close(self):
+        logger.warning('Pool.close() is taking over 60 seconds to complete. '
+                       'Check if you have any unreleased connections left. '
+                       'Use asyncio.wait_for() to set a timeout for '
+                       'Pool.close().')
 
     def terminate(self):
         """Terminate all connections in the pool."""
         if self._closed:
             return
         self._check_init()
-        self._closed = True
         for ch in self._holders:
             ch.terminate()
+        self._closed = True
+
+    async def expire_connections(self):
+        """Expire all currently open connections.
+
+        Cause all currently open connections to get replaced on the
+        next :meth:`~asyncpg.pool.Pool.acquire()` call.
+
+        .. versionadded:: 0.16.0
+        """
+        self._generation += 1
 
     def _check_init(self):
         if not self._initialized:
+            if self._initializing:
+                raise exceptions.InterfaceError(
+                    'pool is being initialized, but not yet ready: '
+                    'likely there is a race between creating a pool and '
+                    'using it')
             raise exceptions.InterfaceError('pool is not initialized')
         if self._closed:
             raise exceptions.InterfaceError('pool is closed')
@@ -549,6 +727,12 @@ class Pool:
             if ch._con is not None:
                 ch._con._drop_local_statement_cache()
 
+    def _drop_type_cache(self):
+        # Drop type codec cache for all connections in the pool.
+        for ch in self._holders:
+            if ch._con is not None:
+                ch._con._drop_local_type_cache()
+
     def __await__(self):
         return self._async__init__().__await__()
 
@@ -637,6 +821,10 @@ def create_pool(dsn=None, *,
         Keyword arguments for the :func:`~asyncpg.connection.connect`
         function.
 
+    :param Connection connection_class:
+        The class to use for connections.  Must be a subclass of
+        :class:`~asyncpg.connection.Connection`.
+
     :param int min_size:
         Number of connection the pool will be initialized with.
 
@@ -688,11 +876,6 @@ def create_pool(dsn=None, *,
        <connection.Connection.add_log_listener>`) present on the connection
        at the moment of its release to the pool.
     """
-    if not issubclass(connection_class, connection.Connection):
-        raise TypeError(
-            'connection_class is expected to be a subclass of '
-            'asyncpg.Connection, got {!r}'.format(connection_class))
-
     return Pool(
         dsn,
         connection_class=connection_class,
